/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package axel;

import Jama.Matrix;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import util.Cluster;
import util.Point;
import util.Util;

/**
 *
 * @author Dunst
 */
//Class represents Gaussian Mixture Model for a spesific set of data points. The covariance matrix is asssumed to be diagonal
public class GaussianMixtureModel implements Serializable{
    //class number
    private int id;
    private int numOfFeatures;
    private int numOfGaussians;
    private List<Cluster> clusters;
    private List<Point> samples;
    //weights per gaussian
    private double[] weights;
    //variance of each feature per gaussian
    private double[][] variances;
    //conditional expectation of the complete loglikelihood (posterior probabilities)
    private double[][] posteriorPrs;
    private double logLikelihood = Double.NaN;
    
    private final double fallbackStdDeviation = 1.0E-8;
    private final double fallbackCovariance = Math.pow(fallbackStdDeviation, 2);
    private final double fallbackGaussianValue = 1.0E-32;
    
    
    public GaussianMixtureModel(int id, List<Point> samples, int numOfGaussians) {
        this.id = id;
        this.numOfGaussians = numOfGaussians;
        this.samples = samples;
        this.numOfFeatures = samples.get(0).size();
    }
    
    public void init(){
        //divide data into given number of subclasses (one subclass - for one gaussian)
        KmeansForGMM kmeans = new KmeansForGMM(samples, numOfGaussians);
        kmeans.initializeCentroidsRandomly();
        kmeans.runClusterer();
        clusters = kmeans.getClusters();
        
        //initialize gaussian parameters
        posteriorPrs = new double[numOfGaussians][samples.size()];
        weights = new double[numOfGaussians];
        for(int i=0; i<numOfGaussians; i++){
            weights[i] = (double)clusters.get(i).size() /samples.size();
        }        
        initMeans();
        initVariances();
    }
    
    private void initMeans(){
        for(int i = 0; i < numOfGaussians; i++){
            clusters.get(i).resetCenter();
        }
    }
    
    private void initVariances(){
        variances = new double[numOfGaussians][numOfFeatures];
        for(int i = 0; i < numOfGaussians; i++){
            double[] variancePerClass = Util.getVariance(clusters.get(i).getSamples(), clusters.get(i).getCenter());        
            for(int j = 0; j < numOfFeatures; j++){
                if(variancePerClass[j] < fallbackCovariance)
                        variances[i][j] = fallbackStdDeviation;
                    else
                        variances[i][j] = Math.sqrt(variancePerClass[j]);
            }
        }
    }
    
    //one iteration for the algorithm
    public void runClusterer(){
        //E step
        //Evaluate the posterior probabilities that relate each cluster to each data point in the conditional probability
        this.calculatePosteriorPrs();
        
        //M step
        //recalculate gaussian parameters on the assumption of calculated previously posterior prs
        List<Point> newMeans = this.recalculateMeans();
        double[][] newVariances = this.recalculateVariances();
        double[] newWeights = this.recalculateWeights();
        
        variances = newVariances;
        weights = newWeights;
        for (int i = 0; i < numOfGaussians; i++){
            clusters.get(i).setCenter(newMeans.get(i));
        }
        
        //calculate loglikelihood
        this.calculateLogLikelihood();
    }
    
    private double[] recalculateWeights(){
        double[] newWeights = new double[numOfGaussians];
        for (int i = 0; i < numOfGaussians; i++){
            newWeights[i] = recalculateWeight(i);
        }
        return newWeights;
    }
    
    private double recalculateWeight(int subclassId){
        double newWeight = 0;
        for(int i = 0; i < this.samples.size(); i++)
            newWeight += this.posteriorPrs[subclassId][i];
        newWeight /= this.samples.size();        
        return newWeight;
    }
    
    private double[][] recalculateVariances(){
        double[][] newVariances = new double[numOfGaussians][numOfFeatures];
        for(int i = 0; i < numOfGaussians; i++)
            newVariances[i] = this.recalculateVariance(i);
        return newVariances;
    }
    
    private double[] recalculateVariance(int subclassId)
    {
        double[] newVariance = new double[numOfFeatures];
        for(int i = 0; i < numOfFeatures; i++){
            newVariance[i] = 0;
        }
        double normFactor = 0;
        
        for(int i = 0; i < clusters.size(); i++) {
            double posteriorPr = this.posteriorPrs[subclassId][i];
            Point mean = clusters.get(subclassId).getCenter();
            Point sample = samples.get(i);            
            for(int j = 0; j < numOfFeatures; j++){
                newVariance[j] = newVariance[j] + Math.pow(sample.get(j)-mean.get(j), 2)*posteriorPr;
            }
            normFactor += posteriorPr;
        }
            
        for(int i = 0; i < numOfFeatures; i++){
            double variance = newVariance[i]/ normFactor;
            
            if(variance < fallbackCovariance)
                newVariance[i] = fallbackStdDeviation;
            else
                newVariance[i] = Math.sqrt(variance);
        }
        return newVariance;
    }
    
    private List<Point> recalculateMeans(){
        List<Point> newMeans = new ArrayList<Point>();
        for(int i = 0; i < numOfGaussians; i++){
            Point newMean = this.recalculateMean(i);
            newMeans.add(newMean);
        }
        return newMeans;
    }
    
    private Point recalculateMean(int subclassId){
        Point mean = new Point();        
        double normFactor = 0;
        for(int i = 0; i < this.samples.size(); i++){
            Point sample = this.samples.get(i);
            double posteriorPr = this.posteriorPrs[subclassId][i];
            mean = Util.addPoints(mean, Util.times(sample, (float)posteriorPr));            
            normFactor += posteriorPr;
        }        
        mean = Util.divideN(mean, (float)normFactor);
        return mean;
    }
    
    //conditional expectation of the complete log likelihood
    private void calculatePosteriorPrs(){
        //calculate normalization factors
        double[] normalizationFactors = new double[samples.size()];
        for(int i = 0; i < samples.size(); i++){
            normalizationFactors[i] = 0;
            for(int j = 0; j < numOfGaussians; j++){
                Point mean = clusters.get(j).getCenter();
                double gaussianValue = weights[j]*Util.getMultivariateGaussian(samples.get(i), mean, variances[j]);
                if(gaussianValue < fallbackGaussianValue)
                    gaussianValue = fallbackGaussianValue;
                normalizationFactors[i] += gaussianValue;
            }
        }
        
        //calculate posterior prs
        for(int i = 0; i < numOfGaussians; i++){
            Point mean = clusters.get(i).getCenter();
            for(int j = 0; j < samples.size(); j++){
                double posteriorPr = weights[i]*Util.getMultivariateGaussian(samples.get(j), mean, variances[i]);
                if(posteriorPr < fallbackGaussianValue)
                    posteriorPr = fallbackGaussianValue;
                posteriorPrs[i][j] = posteriorPr/normalizationFactors[j];
            }
        }
    }
    
    private void calculateLogLikelihood(){
        logLikelihood = 0;
        for(int i = 0; i < samples.size(); i++)
            logLikelihood += Math.log(this.getPdf(samples.get(i)));
    }
    
    public double getPdf(Point sample){
        double pdf = 0;
        for(int i = 0; i < numOfGaussians; i++){
            Point mean = clusters.get(i).getCenter();
            pdf += weights[i] * Util.getMultivariateGaussian(sample, mean, variances[i]);
        }
        if(pdf < fallbackGaussianValue)
            pdf = fallbackGaussianValue;        
        return pdf;
    }

    public int getId() {
        return id;
    }
    
    public double getLogLikelihood(){
        return this.logLikelihood;
    }
}
